#version 150 core
#extension GL_ARB_explicit_attrib_location: enable

in vec2 textureCoord;
in mat3 tbnMatrix;

in vec3 wsPosition;
in vec3 tsPosition;
in vec3 tsCameraPosition;

in float distanceToFragment;

uniform sampler2D textureAlbedo;
uniform sampler2D textureNormal;
uniform sampler2D textureHeight;
uniform sampler2D depthBuffer;

// флаг вывода текста
uniform bool colorFlag;

uniform bool drawPOM;
uniform bool drawSecondGeometryLayer;
// цвет текста
uniform vec3 color;

uniform bool homogeneousClipSpace;
uniform mat4 viewProjectionMatrix;
uniform mat4 projectionInversedMatrix;
uniform mat4 viewInversedMatrix;

uniform vec3 wsCameraPosition;
uniform vec2 screenSize;

// Ouput data
layout( location = 0 ) out vec4 albedoBuffer;
layout( location = 1 ) out vec3 normalBuffer;

// Returns +-1
vec2 signNotZero( in vec2 v ) {
	return vec2( ( v.x >= 0.0 ) ? 1.0 : -1.0, ( v.y >= 0.0 ) ? 1.0 : -1.0 );
}
// Assume normalized input. Output is on [-1, 1] for each component.
vec2 encode( in vec3 v ) {
	// Project the sphere onto the octahedron, and then onto the xy plane
	vec2 p = v.xy * ( 1.0 / ( abs( v.x ) + abs( v.y ) + abs( v.z ) ) );
	// Reflect the folds of the lower hemisphere over the diagonals
	return ( v.z <= 0.0 ) ? ( ( 1.0 - abs( p.yx ) ) * signNotZero( p ) ) : p;
}
vec3 decode( in vec2 e ) {
	vec3 v = vec3( e.xy, 1.0 - abs( e.x ) - abs( e.y ) );
	if ( v.z < 0 ) {
		v.xy = ( 1.0 - abs( v.yx ) ) * signNotZero( v.xy );
	}
	return normalize( v );
}

float getHeight( in vec2 texCoord ) {
	return 1.0f - texture( textureHeight, texCoord ).x;
}
vec2 parallaxOcclusionMapping( in vec2 texCoord, in float currentHeightMapValue ) {
	vec3 cameraDirection  = normalize( tsPosition - tsCameraPosition );
	float heightScale     = 0.01f;
	// number of depth layers
	const float minLayers = 1;
	const float maxLayers = 32;
	int numLayers          = int( mix( maxLayers, minLayers, abs( dot( vec3( 0.0, 0.0, 1.0f ), cameraDirection ) ) ) );
	// calculate the size of each layer
	float layerHeight        = 1.0f / numLayers;
	// depth of current layer
	// the amount to shift the texture coordinates per layer
	vec2 deltaTexCoords      = heightScale * cameraDirection.xy / ( cameraDirection.z * numLayers );
	// vec2 deltaTexCoords      = layerHeight * heightScale * ( cameraDirection.xy / cameraDirection.z );

	// get initial values
	vec2  currentTexCoords      = texCoord;
	float currentLayerHeight    = 0.0f;
	// int i = 0;
	while (
		currentLayerHeight < currentHeightMapValue
		// && i <= ( numLayers + 1 )
	) {
		// get depth of next layer
		currentLayerHeight   += layerHeight;
		// shift texture coordinates along direction of P
		currentTexCoords     -= deltaTexCoords;
		// get depthmap value at current texture coordinates
		currentHeightMapValue = getHeight( currentTexCoords );
		// ++i;
	}
	// if ( i == ( numLayers + 1 ) ) {
	// 	discard;
	// }
	//// now find point via linear interpolation
	// previous point
	vec2 prevTexCoords = currentTexCoords + deltaTexCoords;
	// get depth after and before collision for linear interpolation
	float currentHeight  = currentHeightMapValue - currentLayerHeight;
	float previousHeight = getHeight( prevTexCoords ) - currentLayerHeight + layerHeight;
	// interpolation of texture coordinates
	float weight = currentHeight / ( currentHeight - previousHeight );
	currentTexCoords = prevTexCoords * weight + currentTexCoords * ( 1.0 - weight );

	// interpolation of depth values
	float interpolatedHeight = currentLayerHeight + previousHeight * weight + currentHeight * (1.0 - weight);

	return currentTexCoords;
}

vec3 getNormal( in vec2 texCoord ) {
	vec3 tsNormal = texture( textureNormal, texCoord ).xyz;
	// переводим в NDC (-1.0f, 1.0f)
	vec3 csNormal = normalize( tsNormal * 2.0 - 1.0 );
	// переводим в мировое пространство
	return normalize( tbnMatrix * csNormal );
}
void main() {
	// получаем смещение текстурных координат
	vec2 texCoord = textureCoord;
	float currentHeightMapValue = 0.0f;
	if ( drawPOM && distanceToFragment < 35.0f ) {
		currentHeightMapValue = getHeight( texCoord );
		// POM имеет смысл считать только для фрагментов с ненулевой высотой
		if ( currentHeightMapValue != 0.0f ) {
			texCoord = parallaxOcclusionMapping( texCoord, currentHeightMapValue );
		}
	}

	vec4 colorLinear = texture( textureAlbedo, texCoord );

	if ( colorFlag ) {
		colorLinear = vec4( color, 1.0f );
	} else if ( drawSecondGeometryLayer && drawPOM && distanceToFragment < 35.0f && currentHeightMapValue != 0.0f ) {
		// определяем силуэты для POM
		float deltaCoefficient = 5.0f;
		float positionThreshold = 1.7f;

		vec2 delta = ( texCoord - textureCoord ) * deltaCoefficient;
		// tangent space
		vec3 tsNewPosition = vec3( tsPosition.xy + delta, tsPosition.z );
		// world space
		vec3 wsNewPosition = tbnMatrix * tsNewPosition;
		// clip space
		vec4 csNewPosition = viewProjectionMatrix * vec4( wsNewPosition, 1.0f );
		// [-1,1] NDC
		vec3 ndcNewPosition = csNewPosition.xyz / csNewPosition.w;
		// [0,1] screen space
		vec2 ssNewPosition = ( ndcNewPosition.xy + 1.0f ) * 0.5f;
		// координаты окна
		vec2 windowNewPosition = ssNewPosition * screenSize;
		// глубина нового пикселя
		float newDepth = texelFetch( depthBuffer, ivec2( windowNewPosition ), 0 ).x;
		// если значение есть в буфере (не вышли за границу экрана)
		if ( newDepth != 0.0f ) {
			if ( !homogeneousClipSpace ) {
				// если clip space должен быть в диапазоне [-1, 1]
				newDepth = newDepth * 2.0f - 1.0f;
			}
			// получаем позицию
			vec3 csNewPosition  = vec3( ssNewPosition * 2.0f - 1.0f, newDepth );
			vec4 vsNewPosition  = projectionInversedMatrix * vec4( csNewPosition, 1.0f );
			vsNewPosition      /= vsNewPosition.w;
			vec4 wsNewPosition  = viewInversedMatrix * vsNewPosition;
			// если расстояние между точками значительное
			// и отбрасываемая часть полигона находится перед остальной геометрией, а не позади
			// (векторы из исходного пикселя и из камеры, направленные в новый пиксель - сонаправлены)
			if (
				length( wsPosition - wsNewPosition.xyz ) > positionThreshold &&
			    dot( wsNewPosition.xyz - wsPosition, wsNewPosition.xyz - wsCameraPosition ) > 0
			) {
				// отбрасываем пиксель
				colorLinear = vec4( 0.0f, 0.0f, 0.0f, 0.0f );
			}
		}
	}

	albedoBuffer    = colorLinear;
	//normalBuffer.xy = encode( getNormal( texCoord ) );
	normalBuffer = getNormal( texCoord );
}


