#version 150 core

uniform sampler2D depthBuffer;
uniform sampler2D normalBuffer;
uniform samplerCubeShadow shadowBuffer;


uniform mat4 lightProjectionMatrix;

// для восстановления позиции из глубины
uniform bool homogeneousClipSpace;
uniform mat4 projectionInversedMatrix;
uniform mat4 viewInversedMatrix;

// для преобразования вектора в пространстве мира в глубину в пространстве света
//uniform vec2 lightProjectionInfo;
uniform bool softShadows;


uniform vec2 screenSize;

uniform bool drawLights;
uniform bool drawShadows;

uniform int lightType;
uniform vec3 wsLightPosition;
uniform vec3 lightColor;
// uniform float lightBrightness;
uniform float lightRadius;
uniform float lightAttenuationConstant;
uniform float lightAttenuationLinear;
uniform float lightAttenuationQuadratic;
uniform vec3 wsLightDirection;
uniform float lightConeInner;
uniform float lightConeOuter;


out vec3 fragmentColor;


//Sphere of samples for the World Space sampling
const vec3 poissonSphere[ 16 ] = vec3[](
	vec3( 0.53812504, 0.18565957, -0.43192 ),
	vec3( 0.13790712, 0.24864247, 0.44301823 ),
	vec3( 0.33715037, 0.56794053, -0.005789503 ),
	vec3( -0.6999805, -0.04511441, -0.0019965635 ),
	vec3( 0.06896307, -0.15983082, -0.85477847 ),
	vec3( 0.056099437, 0.006954967, -0.1843352 ),
	vec3( -0.014653638, 0.14027752, 0.0762037 ),
	vec3( 0.010019933, -0.1924225, -0.034443386 ),
	vec3( -0.35775623, -0.5301969, -0.43581226 ),
	vec3( -0.3169221, 0.106360726, 0.015860917 ),
	vec3( 0.010350345, -0.58698344, 0.0046293875 ),
	vec3( -0.08972908, -0.49408212, 0.3287904 ),
	vec3( 0.7119986, -0.0154690035, -0.09183723 ),
	vec3( -0.053382345, 0.059675813, -0.5411899 ),
	vec3( 0.035267662, -0.063188605, 0.54602677 ),
	vec3( -0.47761092, 0.2847911, -0.0271716 )
);

vec2 getSSPosition( in vec2 windowSpacePosition ) {
	return windowSpacePosition / screenSize;
}
/** Read the world space position of the point at screen-space pixel windowSpacePosition */
vec4 getWSPosition( in vec2 windowSpacePosition, in float z ) {
	if ( !homogeneousClipSpace ) {
		// если clip space должен быть в диапазоне [-1, 1]
		z = z * 2.0f - 1.0f;
	}
	vec3 csPosition  = vec3( getSSPosition( windowSpacePosition ) * 2.0f - 1.0f, z );
	vec4 vsPosition  = projectionInversedMatrix * vec4( csPosition, 1.0f );
	vsPosition      /= vsPosition.w;
	vec4 wsPosition  = viewInversedMatrix * vsPosition;
	return wsPosition;
}
vec4 getWSPositionFromDepth( in vec2 windowSpacePosition ) {
	float z = texelFetch( depthBuffer, ivec2( windowSpacePosition ), 0 ).x;
	return getWSPosition( windowSpacePosition, z );
}


// Let V be the light-to-fragment direction vector.
// V needs to be in world space since the shadow cube map is normally axis aligned in world space
// thus the "max(abs(V.x),abs(V.y),abs(V.z))"-part only works if V is a world space direction vector.
// The Z value in the respective cube side frustum/"eye space" can be calculated by taking the max of the absolute values of V's components.
// Z = max(abs(V.x),abs(V.y),abs(V.z))
// Then, to be precise, we should negate Z because in OpenGL, the negative Z-axis points into the screen/view frustum.
// Now we want to get the depth buffer "compatible" value of that -Z.
// Looking at the OpenGL perspective matrix we see that, for any homogeneous vector multiplied with that matrix,
// the resulting z value is completely independent of the vector's x and y components.
// So we can simply multiply this matrix with the homogeneous vector (0,0,-Z,1) and we get the vector (components):
// x = 0
// y = 0
// z = (-Z * -(f+n) / (f-n)) + (-2*f*n / (f-n))
// w = Z
// Then we need to do the perspective divide, so we divide z by w (Z) which gives us:
// z' = (f+n) / (f-n) - 2*f*n / (Z* (f-n))
// This z' is in OpenGL's normalized device coordinate (NDC) range [-1,1] and needs to be transformed into a depth buffer compatible range of [0,1]:
// z_depth_buffer_compatible = (z' + 1.0) * 0.5
float vectorToDepth( in vec3 fragmentToLight ) {
	vec3 absVector = abs( fragmentToLight );
	float lsDepth  = max( absVector.x, max( absVector.y, absVector.z ) );

	vec4 homogeneousVector = lightProjectionMatrix * vec4( 0, 0, -lsDepth, 1 );
	// NDC depth
	float csDepth = homogeneousVector.z / homogeneousVector.w;

	if ( !homogeneousClipSpace ) {
		// если clip space в диапазоне [-1, 1], преобразуем в [0, 1]
		return ( csDepth + 1.0f ) * 0.5f;
	}
	return csDepth;
}
float calculateShadow( in vec3 lightDirection ) {
	float shadow = 0.0f;
	if ( softShadows ) {
		// PCF
		int samples      = 16;
		float diskRadius = 0.05f;

		for ( int i = 0; i < samples; ++i ) {
			vec3 newLightDirection = lightDirection + poissonSphere[ i ] * diskRadius;
			float fragmentDepth    = vectorToDepth( newLightDirection );
			shadow                += texture( shadowBuffer, vec4( newLightDirection, fragmentDepth ) );
		}
		shadow /= float(samples);
	} else {
		// simple shadow
		float fragmentDepth = vectorToDepth( lightDirection );
		shadow              = texture( shadowBuffer, vec4( lightDirection, fragmentDepth ) );
	}
	return shadow;
}



// Затухание
float calculateAttenuation( in float lightDistance ) {
	// расчитываем затухание только в пределах радиуса
	float d         = min( lightRadius, lightDistance );
	float nominator = clamp( 1.0f - pow( d / lightRadius, 4 ), 0.0f, 1.0f );
	// делимое в квадрате для более плавного затухания + учет яркости источника
	nominator = nominator * nominator /* lightBrightness*/;
	// квадратичное затухание с параметрами
	float denominator = lightAttenuationConstant + lightAttenuationLinear * d + lightAttenuationQuadratic * d * d;
	return nominator / denominator;
}

vec3 calculateLight( in vec3 wsNormal, in vec3 lightDirection, in float lightDistance ) {
	vec3 fragmentDirection = normalize( -lightDirection );
	// рассеянный свет
	vec3 diffuse = max( dot( wsNormal, fragmentDirection ), 0.0f ) * lightColor;
	if ( drawLights ) {
		// если прожектор
		if ( lightType == 1 ) {
			float theta     = dot( fragmentDirection, normalize( -wsLightDirection ) );
			float epsilon   = lightConeInner - lightConeOuter;
			float intensity = clamp( ( theta - lightConeOuter ) / epsilon, 0.0f, 1.0f );
			diffuse        *= intensity;
		}
		// по умолчанию тени нет
		float shadowCoefficient = 1.0f;
		// TODO: сделать поддержку спотов
		if ( drawShadows && lightType == 0 ) {
			// тень
			shadowCoefficient = calculateShadow( lightDirection );
			diffuse          *= shadowCoefficient;
		}
		// считаем затухание только для участков, находящихся не в тени
		if ( shadowCoefficient != 0.0f ) {
			diffuse *= calculateAttenuation( lightDistance );
		}

	}
	return diffuse;
}

void main() {
	vec3 wsPosition     = getWSPositionFromDepth( gl_FragCoord.xy ).xyz;
	vec3 lightDirection = wsPosition - wsLightPosition;
	float lightDistance = length( lightDirection );
	// считаем только те фрагменты, которые находятся на освещаемой источником поверхности
	if ( lightDistance > lightRadius ) {
		discard;
	}
	vec3 wsNormal = texelFetch( normalBuffer, ivec2( gl_FragCoord.xy ), 0 ).xyz;
	// и повернуты к нему лицом
	if ( dot( wsNormal, -lightDirection ) <= 0.0f ) {
		discard;
	}
	fragmentColor = calculateLight( wsNormal, lightDirection, lightDistance );
}
