#version 150 core

uniform sampler2D normalBuffer;
uniform sampler2D depthFirstBuffer;
uniform sampler2D depthSecondBuffer;

uniform vec2 screenSize;
uniform vec3 wsCameraPosition;

uniform bool homogeneousClipSpace;
uniform mat4 projectionInversedMatrix;
uniform mat4 viewInversedMatrix;

uniform bool drawSecondGeometryLayer;

uniform int numSamples;
uniform int numSpiralTurns;
uniform int logMaxOffset;
uniform int maxMipLevel;

/** The height in pixels of a 1m object if viewed from 1m away.
	You can compute it from your projection matrix.  The actual value is just
	a scale factor on radius; you can simply hardcode this to a constant (~500)
	and make your radius value unitless (...but resolution dependent.)  */
uniform float projScale;

/** World-space AO radius in scene units (r).  e.g., 1.0m */
uniform float radius;
uniform float radius2;
uniform float invRadius2;

/** Bias to avoid AO in smooth corners, e.g., 0.01m */
uniform float bias;

/** intensity / radius^6 */
uniform float intensity;
uniform float intensityDivR6;

out float visibility;


vec2 getSSPosition( in vec2 windowSpacePosition ) {
	return windowSpacePosition / screenSize;
}
/** Read the world space position of the point at screen-space pixel windowSpacePosition */
vec4 getWSPosition( in vec2 windowSpacePosition, in float z ) {
	if ( !homogeneousClipSpace ) {
		// если clip space должен быть в диапазоне [-1, 1]
		z = z * 2.0f - 1.0f;
	}
	vec3 csPosition  = vec3( getSSPosition( windowSpacePosition ) * 2.0f - 1.0f, z );
	vec4 vsPosition  = projectionInversedMatrix * vec4( csPosition, 1.0f );
	vsPosition      /= vsPosition.w;
	vec4 wsPosition  = viewInversedMatrix * vsPosition;
	return wsPosition;
}
vec4 getWSPositionFromDepth( in vec2 windowSpacePosition ) {
	float z = texelFetch( depthFirstBuffer, ivec2( windowSpacePosition ), 0 ).x;
	return getWSPosition( windowSpacePosition, z );
}



/** Returns a unit vector and a screen-space radius for the tap on a unit disk
	(the caller should scale by the actual disk radius) */
vec2 tapLocation(int sampleNumber, float spinAngle, out float ssR){
	// Radius relative to ssR
	float alpha = float(sampleNumber + 0.5) * (1.0 / numSamples);
	float angle = alpha * ( numSpiralTurns * 6.28f ) + spinAngle;

	ssR = alpha;
	return vec2(cos(angle), sin(angle));
}

int getMipLevel(float ssR) {
	// Derivation:
	//  mipLevel = floor(log(ssR / MAX_OFFSET));
	return clamp( int( floor( log2( ssR ) ) ) - logMaxOffset, 0, maxMipLevel );
}

/** Read the camera-space position of the point at screen-space pixel ssP + unitOffset * ssR.  Assumes length(unitOffset) == 1. */
void getOffsetPositions(ivec2 ssCoord, vec2 unitOffset, float ssR, out vec3 P1, out vec3 P2 ) {
	int mipLevel = getMipLevel(ssR);
	ivec2 ssP = ivec2(ssR * unitOffset) + ssCoord;

	// We need to divide by 2^mipLevel to read the appropriately scaled coordinate from a MIP-map.
	// Manually clamp to the texture size because texelFetch bypasses the texture unit
	ivec2 mipP = clamp(ssP >> mipLevel, ivec2(0), textureSize( depthFirstBuffer, mipLevel ) - ivec2(1));

	P1.z = texelFetch( depthFirstBuffer, mipP, mipLevel ).r;
	// Offset to pixel center
	P1 = getWSPosition( vec2( ssP ) + 0.5, P1.z ).xyz;

	if ( drawSecondGeometryLayer ) {
		P2.z = texelFetch( depthSecondBuffer, mipP, mipLevel ).r;
		// Offset to pixel center
		P2 = getWSPosition( vec2( ssP ) + 0.5, P2.z ).xyz;
	}
}


float fallOffFunction(float vv, float vn, float epsilon) {
	// A: From the HPG12 paper
	// Note large epsilon to avoid overdarkening within cracks
	//  Assumes the desired result is intensity/radius^6 in main()
	// return float(vv < radius2) * max((vn - bias) / (epsilon + vv), 0.0) * radius2 * 0.6;

	// B: Smoother transition to zero (lowers contrast, smoothing out corners). [Recommended]
	// Epsilon inside the sqrt for rsqrt operation
	float f = max(1.0 - vv * invRadius2, 0.0); return f * max((vn - bias) * inversesqrt(epsilon + vv), 0.0);
	// LOW QUALITY: Avoid the square root from above.
    //  Assumes the desired result is intensity/radius^6 in main()
	// float f = max(radius2 - vv, 0.0); return f * f * f * max((vn - bias) / (epsilon + vv), 0.0);

	// C: Medium contrast (which looks better at high radii), no division.  Note that the
	// contribution still falls off with radius^2, but we've adjusted the rate in a way that is
	// more computationally efficient and happens to be aesthetically pleasing.  Assumes
	// division by radius^6 in main()
	// return 4.0 * max(1.0 - vv * invRadius2, 0.0) * max(vn - bias, 0.0);

	// D: Low contrast, no division operation
	//return 2.0 * float(vv < radius * radius) * max(vn - bias, 0.0);
}


/** Compute the occlusion due to sample point \a Q about camera-space point \a wsPosition with unit normal \a wsNormal */
float aoValueFromPositionsAndNormal(vec3 wsPosition, vec3 wsNormal, vec3 Q) {
	vec3 v = Q - wsPosition;
	float vv = dot(v, v);
	float vn = dot(v, wsNormal);
	const float epsilon = 0.001;

	// Without the angular adjustment term, surfaces seen head on have less AO
	return fallOffFunction(vv, vn, epsilon) * mix(1.0, max(0.0, 1.5 * wsNormal.z), 0.35);
}

/** Compute the occlusion due to sample with index \a i about the pixel at \a ssCoord that corresponds
	to camera-space point \a wsPosition with unit normal \a wsNormal, using maximum screen-space sampling radius \a ssDiskRadius

	Note that units of H() in the HPG12 paper are meters, not
	unitless.  The whole falloff/sampling function is therefore
	unitless.  In this implementation, we factor out (9 / radius).

	Four versions of the falloff function are implemented below

	When sampling from the peeled depth buffer, make sure ssDiskRadius has been premultiplied by cszBufferScale
*/
float sampleAO(in ivec2 ssCoord, in vec3 wsPosition, in vec3 wsNormal, in float ssDiskRadius, in int tapIndex, in float randomPatternRotationAngle ) {
	// Offset on the unit disk, spun for this pixel
	float ssR;
	vec2 unitOffset = tapLocation(tapIndex, randomPatternRotationAngle, ssR);

	// Ensure that the taps are at least 1 pixel away
	ssR = max(0.75, ssR * ssDiskRadius);

	// The occluding point in camera space
	vec3 Q1;
	vec3 Q2;
	getOffsetPositions(ssCoord, unitOffset, ssR, Q1, Q2 );

	if ( drawSecondGeometryLayer ) {
		return max(aoValueFromPositionsAndNormal(wsPosition, wsNormal, Q1), aoValueFromPositionsAndNormal(wsPosition, wsNormal, Q2));
	} else {
		return aoValueFromPositionsAndNormal(wsPosition, wsNormal, Q1);
	}
}



const float MIN_RADIUS = 3.0f; // pixels

void main() {
	// Pixel being shaded
	ivec2 ssCoord = ivec2( gl_FragCoord.xy );

	// World space point being shaded
	vec3 wsPosition = getWSPositionFromDepth( gl_FragCoord.xy ).xyz;
	// расстояние от камеры до фрагмента сцены
	float distanceToFragment = length( wsPosition - wsCameraPosition );
	// Choose the screen-space sample radius
	// proportional to the projected area of the sphere
	float ssDiskRadius = projScale * radius / distanceToFragment;
	if (ssDiskRadius <= MIN_RADIUS || distanceToFragment < radius + 0.3f ) {
		// There is no way to compute AO at this radius
		visibility = 1.0;
		return;
	}

	vec3 wsNormal = texelFetch( normalBuffer, ssCoord, 0 ).xyz;
	// Hash function used in the HPG12 AlchemyAO paper
	float randomPatternRotationAngle = (((3 * ssCoord.x) ^ (ssCoord.y + ssCoord.x * ssCoord.y))	) * 10;
	float sum = 0.0;
	for (int i = 0; i < numSamples; ++i) {
		sum += sampleAO(ssCoord, wsPosition, wsNormal, ssDiskRadius, i, randomPatternRotationAngle);
	}

	float A = pow(max(0.0, 1.0 - sqrt(sum * (3.0 / numSamples))), intensity);
	// // LOW QUALITY
	// float A = max(0.0, 1.0 - sum * intensityDivR6 * (5.0 / numSamples));
	// // Anti-tone map to reduce contrast and drag dark region farther
	// // (x^0.2 + 1.2 * x^4)/2.2
	// A = (pow(A, 0.2) + 1.2 * A*A*A*A) / 2.2;

	// Fade in as the radius reaches 2 pixels
	visibility = mix(1.0, A, clamp( ( ssDiskRadius - MIN_RADIUS ), 0.0, 1.0 ) );

	// visibility = 1.0f;
}

